// Tony Urschler
// 10/11/21

#include <limits.h>
#include <signal.h>
#include <stdlib.h>
#include <string.h>
#include <sys/shm.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <time.h>
#include <unistd.h>
#include "config.h"
#include "liblicense.h"

// TODO: timer doesn't kill exec process

// globals
int segment_id;
int* nlicenses;
int semids[2];
pid_t pids[MAX_FORKS];
int numofchldrn;
int totchldrn;
pid_t gchild;

// prototypes
void killandfree(void);
void handle_sigint(int sig);
void child_handle_sigint(int sig);
static void handle_timer(int sig);
static int setupinterrupt(void);
static int setuptimer(int time);
void docommand(char* cline);

int main(int argc, char* argv[]) {
    
    // setting error message
    char errmsgraw[100];
    snprintf(errmsgraw, sizeof(errmsgraw), "%s: Error", argv[0]);
    char* errmsg = errmsgraw+2;

    // handling sigint
    signal(SIGINT, handle_sigint);

	int c;
    int tflag = 0;
    char* tvalue = NULL;
    char* usagestmt = "Usage: runsim [-t sec] n";
    
    // reading and validating arguments
    while ((c = getopt(argc, argv, "ht:")) != -1) {
        switch (c) {
            case 'h':
                printf("%s\n", usagestmt);
                exit(0);
            case 't':
                tflag = 1;
                tvalue = optarg;
                break;
            default:
                exit(EXIT_FAILURE);
        }
    }
    if (argc < 2) {
        printf("%s\n", usagestmt);
        exit(EXIT_FAILURE);
    }
    
    int n, time;
    
    // validating n
    if (argv[optind] != NULL) {
        n = atoi(argv[optind]);
        if (n > 20) {
            n = 20;
            printf("Warning: Max number of processes at a time is 20.\n\t n is set to 20.\n");
        }
        else if (n < 1) {
            errno = EINVAL;
            perror(errmsg);
            exit(EXIT_FAILURE);
        }
    }
    else {
        printf("%s\n", usagestmt);
        exit(EXIT_FAILURE);
    }
    
    // storing time
    if (tflag > 0)
        time = atoi(tvalue);
    else
        time = TIME_LIMIT;
    
    // setting up timer
    if (setupinterrupt() == -1) {
        snprintf(errmsgraw, sizeof(errmsgraw),
                errmsg, ": Failed to setup handler for SIGPROF");
        char* setuperr = errmsgraw;
        perror(setuperr);
        exit(EXIT_FAILURE);
    }
    if (setuptimer(time) == -1) {
        snprintf(errmsgraw, sizeof(errmsgraw),
                errmsg, ": Failed to setup ITIMER_PROF interval timer");
        char* setuperr = errmsgraw;
        perror(setuperr);
        exit(EXIT_FAILURE);
    }
    
    // allocating shared memory
    segment_id = shmget(IPC_PRIVATE, sizeof(int),
            IPC_CREAT | IPC_EXCL | S_IRUSR | S_IWUSR);
    nlicenses = (int*) shmat(segment_id, NULL, 0);
    *nlicenses = n;

    char str[MAX_CANON];
    numofchldrn = 0;
    totchldrn = 0;
    int status = 0;
    pid_t termpid;
    int x;
    
    // setup semaphore
    x = initlicense(n, semids);
    if (x == -1) {
        snprintf(errmsgraw, sizeof(errmsgraw),
                errmsg, ": License failed to initialize");
        char* initerr = errmsgraw;
        perror(initerr);
        shmdt(nlicenses);
        shmctl(segment_id, IPC_RMID, 0);
        exit(EXIT_FAILURE);
    }
    
    // fgets until EOF
    while (fgets(str, MAX_CANON, stdin) != NULL) {
        // make child
        pids[numofchldrn] = fork();

        // if child, call docommand
        if (pids[numofchldrn] == 0) {
            signal(SIGINT, child_handle_sigint);
            strtok(str, "\n");
            docommand(str);
        }
        else {
            totchldrn++;
            numofchldrn++;
            
            // check if anyone died
            if ((termpid = waitpid(-1, &status, WNOHANG)) > 0) {
                printf("%d has died\n", termpid);
                x = returnlicense(semids[0], nlicenses);
                if (x == -1) {
                    snprintf(errmsgraw, sizeof(errmsgraw),
                            errmsg, ": Failed to return license");
                    char* reterr = errmsgraw;
                    perror(reterr);
                    killandfree();
                    exit(EXIT_FAILURE);
                }
            }

            // don't make more than MAX_FORKS children
            if (numofchldrn >= MAX_FORKS)
                break;
        }
    }
    numofchldrn--;

    // wait until everyone has died
    do {
        termpid = wait(&status);
        x = returnlicense(semids[0], nlicenses);
        if (x == -1) {
            snprintf(errmsgraw, sizeof(errmsgraw),
                    errmsg, ": Failed to return license");
            char* reterr = errmsgraw;
            perror(reterr);
            killandfree();
            exit(EXIT_FAILURE);
        }
    } while (termpid > 0);
    
    printf("All processes complete.\n");
    killandfree();

    exit(EXIT_SUCCESS);
}


/* *** Functions *** */


// gets a license and forks a child that calls execlp
void docommand(char* cline) {
    // wait for license
    if (getlicense(semids[0], nlicenses) == -1) {
        exit(EXIT_FAILURE);
    }
    
    // breakup string to get args
    char* word1 = strtok(cline, " ");
    word1 = strtok(NULL, " ");
    char* word2 = strtok(NULL, " ");
    char buff[10];
    sprintf(buff, "%d", semids[1]);
    char* word3 = buff;
    char* progName = "./testsim";

    // fork child
    pid_t term;
    int status;
    gchild = fork();
    if (gchild == 0)
        execlp(progName, progName, word1, word2, word3, NULL);
    else
        term = wait(&status);
    exit(EXIT_SUCCESS);
}

// kills all child processes and frees semaphores/shm
void killandfree() {
    for (int i = 0; i < numofchldrn+1; i++) {
        kill(pids[i], SIGINT);
    }

    // close and unlink semaphore
    int rc = semctl(semids[0], 1, IPC_RMID);
    int rc1 = semctl(semids[1], 1, IPC_RMID);
    if (rc < 0 || rc1 < 0) {
        exit(EXIT_FAILURE);
    }

    // deallocating shared memory
    shmdt(nlicenses);
    shmctl(segment_id, IPC_RMID, 0);
    
    // writing terminating msg to file
    time_t now;
    time(&now);
    struct tm* tp = localtime(&now);
    char* msg; char buff[50];
    sprintf(buff, "TERMINATED AT: %.2d:%.2d:%.2d",
            tp->tm_hour, tp->tm_min, tp->tm_sec);
    msg = buff;

    FILE* fptr;
    fptr = fopen("logfile", "a");
    fprintf(fptr, "%s\n", msg);
    fclose(fptr); 

    exit(EXIT_SUCCESS);
}

// handles sigint
void handle_sigint(int sig) {
    printf("\nSIGINT caught! Killing children and freeing memory\n");
    killandfree();
}

// called when the timer interval has occurred
void handle_timer(int sig) {
    printf("Times up! Killing children and freeing shared memory.\n");
    killandfree();
}

// sets up the interrupt for timer interval
static int setupinterrupt(void) {
    struct sigaction act;
    act.sa_handler = handle_timer;
    act.sa_flags = 0;
    return (sigemptyset(&act.sa_mask) || sigaction(SIGALRM, &act, NULL));
}

// sets up timer
static int setuptimer(int time) {
    struct itimerval value;
    value.it_interval.tv_sec = time;
    value.it_interval.tv_usec = 0;
    value.it_value = value.it_interval;
    return (setitimer(ITIMER_REAL, &value, NULL));
}

// child sigint
void child_handle_sigint(int sig) {
    kill(gchild, SIGINT);
    exit(EXIT_SUCCESS);
}


